/*
** identd.c                       A TCP/IP link identification protocol server
**
** Version: 1.8.1
**
** This program is in the public domain and may be used freely by anyone
** who wants to. 
**
** Last update: 23 May 1992
**
** Please send bug fixes/bug reports to: Peter Eriksson <pen@lysator.liu.se>
*/

#define VERSION "[in.identd, version 1.8.1]"

#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <netdb.h>
#include <nlist.h>
#include <pwd.h>
#include <signal.h>
#include <syslog.h>

#ifdef HAVE_KVM
#include <kvm.h>
#else
#include "kvm.h"
#endif

#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/ioctl.h>
#define KERNEL
#include <sys/user.h>
#include <sys/proc.h>
#include <sys/protosw.h>
#include <sys/domain.h>
#undef KERNEL

#ifndef HPUX7
#define KERNEL
#endif

#include <sys/file.h>

#ifndef HPUX7
#undef KERNEL
#endif

#include <fcntl.h>

#ifdef ultrix
#include <sys/dir.h>
#endif

#include <sys/user.h>
  
#include <net/if.h>
#include <net/route.h>
#include <netinet/in.h>
#include <netinet/in_pcb.h>
#include <netinet/tcp.h>
#include <netinet/ip_var.h>
#include <netinet/tcp_timer.h>
#include <netinet/tcp_var.h>
#include <arpa/inet.h>

#ifdef MIPS
#include <sysv/sys/var.h>
extern int errno;
#endif


extern void *calloc();


struct nlist nl[] =
{
#ifdef hp9000s800	/* HP9000 Series 800, HP-UX 8.0 */

#define N_FILE  0  
#define N_NFILE 1
#define N_TCB   2

  { "file" },
  { "nfile" },
  { "tcb" },
  
#else
#ifdef HPUX7		/* HP-UX 7.0 */

#define N_FILE  0  
#define N_NFILE 1
#define N_TCB   2

  { "_file" },
  { "_nfile" },
  { "_tcp_cb" },

#else
#ifdef MIPS		/* MIPS RISC/OS */

#define N_FILE 0  
#define N_V    1
#define N_TCB  2

  { "_file" },
  { "_v" },
  { "_tcb" },

#else			/* SUN or BSD or other systems */

#define N_FILE 0  
#define N_NFILE 1
#define N_TCB 2
#define N_PROC 3
#define N_NPROC 4
#define	N_QS	5

  { "_file" },
  { "_nfile" },
  { "_tcb" },
  { "_proc" },
  { "_nproc" },
  { "_qs" },

#endif
#endif
#endif
  { "" }
};


char *path_unix = NULL;
char *path_kmem = NULL;

kvm_t *kd;

int verbose_flag = 1;
int debug_flag = 1;
int syslog_flag = 0;

#ifdef LOG_DAEMON
int syslog_facility = LOG_DAEMON;
#endif

struct file *xfile;
int nfile;

#ifdef MIPS
struct var v;
#endif

int lport, fport;
struct inpcb tcb;


/*
** Error handling macros
*/
#define ERROR(fmt) \
    ((syslog_flag ? syslog(LOG_ERR, fmt) : 0), \
     (debug_flag ? (fprintf(stderr, "%d, %d: ERROR: X-DBG", lport, fport), \
		    fprintf(stderr, fmt), perror(""), 0) : \
      (printf("%d, %d: ERROR: UNKNOWN-ERROR\r\n", lport, fport), \
       exit(1), 0)))

#define ERROR1(fmt,v1) \
    ((syslog_flag ? syslog(LOG_ERR, fmt, v1) : 0), \
     (debug_flag ? (fprintf(stderr, "%d, %d: ERROR: X-DBG: ", lport, fport), \
		    fprintf(stderr, fmt, v1), perror(""), 0) : \
      (printf("%d, %d: ERROR: UNKNOWN-ERROR\r\n", lport, fport), \
       exit(1), 0)))

#define ERROR2(fmt,v1,v2) \
    ((syslog_flag ? syslog(LOG_ERR, fmt, v1, v2) : 0), \
     (debug_flag ? (fprintf(stderr, "%d, %d: ERROR: X-DBG: ", lport, fport), \
		    fprintf(stderr, fmt, v1, v2), perror(""), 0) : \
      (printf("%d, %d: ERROR: UNKNOWN-ERROR\r\n", lport, fport), \
       exit(1), 0)))


/*
** The structure passing convention for GCC is incompatible with
** Suns own C compiler, so we define our own inet_ntoa() function.
** (This should only affect GCC version 1 I think, a well, this works
** for version 2 also so why bother.. :-)
*/
#if defined(__GNUC__) && defined(__sparc__)

#ifdef inet_ntoa
#undef inet_ntoa
#endif

char *inet_ntoa(ad)
  struct in_addr ad;
{
  unsigned long int s_ad;
  int a, b, c, d;
  static char addr[20];
  
  s_ad = ad.s_addr;
  d = s_ad % 256;
  s_ad /= 256;
  c = s_ad % 256;
  s_ad /= 256;
  b = s_ad % 256;
  a = s_ad / 256;
  sprintf(addr, "%d.%d.%d.%d", a, b, c, d);
  
  return addr;
}
#endif


/*
** Return the name of the connecting host, or the IP number as a string.
*/
char *gethost(addr)
  struct in_addr *addr;
{
  struct hostent *hp;

  
  hp = gethostbyaddr(addr, sizeof(struct in_addr), AF_INET);
  if (hp)
    return hp->h_name;
  else
    return inet_ntoa(*addr);
}


/*
** Get a piece of kernel memory with error handling.
** Always return 1.
*/
int getbuf(addr, buf, len, what)
  long addr;
  char *buf;
  int len;
  char *what;
{

  if (kvm_read(kd, addr, buf, len) < 0)
    ERROR2("getbuf: kvm_read(%08x, %d)", addr, len);
  
  return 1;
}



int main(argc,argv)
  int argc;
  char *argv[];
{
  int i, j;
  struct proc process;
  struct user user_info;
  long nextaddr;
  int nproc;
  struct prochd qs[NQS];
  struct file **proc_files;
  struct file this_file;
  struct socket sockinfo;
  struct protosw protoinfo;
  struct domain domaininfo;
  struct inpcb tsinfo;
  int nfile;
  int	*Counted;
  struct file *file_list;
  int total_connects;
  struct in_addr remaddr;
  int port;
  struct passwd *pe;
  struct sess session;
  int local = 0;


  if (argc != 3)
  {
	fprintf(stderr, "Usage: %s remote_ip_address remote_port\n",
		argv[0]);
	exit(1);
  }
  if (!strcmp(argv[1], "local"))
	local = 1;
  else
	remaddr.s_addr = inet_addr(argv[1]);
  port = atoi(argv[2]);
  if (!(kd = kvm_open(path_unix, path_kmem, NULL, O_RDONLY, NULL)))
    ERROR("main: kvm_open");
  if (kvm_nlist(kd, nl) != 0)
    ERROR("main: kvm_nlist");

  printf("USER         PID CONNs\n");
  /* Find the process table and look through it */
  getbuf(nl[N_NFILE].n_value, &nfile, sizeof(nfile), NULL);
  Counted = malloc(sizeof(int) * nfile);
  getbuf(nl[N_FILE].n_value, &file_list, sizeof(file_list), NULL);
  getbuf(nl[N_NPROC].n_value, &nproc, sizeof(nproc), NULL);
  getbuf(nl[N_PROC].n_value, &nextaddr, sizeof(nextaddr), NULL);
  for (i = 0; i < nproc; i++)
  {
    getbuf(nextaddr, &process, sizeof(process), NULL);
    if (process.p_pid > 0 &&
	process.p_stat != SZOMB &&
	process.p_stat != SIDL)
    {
      /* Get the user structure */
      getbuf(process.p_uarea, &user_info, sizeof(user_info), NULL);
      proc_files = malloc((user_info.u_lastfile+1) * sizeof(struct file *));
      getbuf(user_info.u_ofile, proc_files,
		(user_info.u_lastfile+1) * sizeof(struct file *),
		NULL);
      for (i = 0; i < nfile; i++)
	Counted[i] = 0;
      total_connects = 0;
      for (i = 0; i <= user_info.u_lastfile; i++)
      {
	if (!proc_files[i])
	    continue;
	j = proc_files[i] - file_list;
	if (Counted[j])
	   continue;
	Counted[j] = 1;
	getbuf(proc_files[i], &this_file, sizeof(this_file), NULL);
	if (this_file.f_type == DTYPE_SOCKET)
	{
		/* We know it has a socket, now lets find out about it */
		getbuf(this_file.f_data, &sockinfo, sizeof(sockinfo), NULL);
		if (sockinfo.so_type != SOCK_STREAM)
			continue;
		if (!(sockinfo.so_state&SS_ISCONNECTED))
			continue; /* Don't care about these */
		getbuf(sockinfo.so_proto, &protoinfo, sizeof(protoinfo), NULL);
		getbuf(protoinfo.pr_domain, &domaininfo, sizeof(domaininfo), NULL);
		if (domaininfo.dom_family != AF_INET)
			continue;
		getbuf(sockinfo.so_pcb, &tsinfo, sizeof(tsinfo), NULL);
		/* Finally! */
		if ((local &&tsinfo.inp_lport == port) ||
		    (!local && tsinfo.inp_fport == port &&
			tsinfo.inp_faddr.s_addr == remaddr.s_addr))
		{
			total_connects++;
		}
	}
      }
      if (total_connects)
      {
	pe = getpwuid((int) process.p_uid);
	if (pe)
	  printf("%-10s %5d %3d\n",
		pe->pw_name,
		(unsigned) (unsigned short) process.p_pid,
		total_connects);
	else
	  printf("%-10d %5d %3d\n",
		(int) process.p_uid,
		(unsigned) (unsigned short) process.p_pid,
		total_connects);
      }
      free(proc_files);
    }
    nextaddr += sizeof(process);
  }
  exit(0);
  }
